/*
** Copyright (c) 2000-2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: nk_snmp.c,v 1.1.1.1 2001/12/10 05:18:33 bambi Exp $
**
*/

/*
** Module	: nkSnmp
** Purpose	: An abstraction layer between NK and the underlying
**		  SNMP implementation.  The SNMP library can be changed
**		  without impacting upon the entire application.  Only
**		  this module will need to be rewritten.
** Exports	: 
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <string.h>

#include "snmp/config.h"
#include "asn1.h"
#include "snmp.h"
#define SNMP_NEED_REQUEST_LIST
#include "snmp_api.h"
#include "snmp_client.h"
#include "snmp_impl.h"
#include "parse.h"
#include "mib.h"
#include "system.h"
#include "int64.h"
#include "snmpv3.h"    
#include "read_config.h"
#include "snmp_debug.h"
#include "callback.h"
#include "snmpusm.h"
#include "tools.h"
#include "keytools.h"
#include "lcd_time.h"
#include "snmp_alarm.h"
#include "snmp_logging.h"
#include "default_store.h"
#include "mt_support.h"
#include "snmp-tc.h"
#include "snmp_parse_args.h"



#include "nk_common.h"
#include "nk_snmp.h"

#define PACKET_LENGTH 4500


/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/

static	int errorCode;
static	char errorMsg[1024];

u_char *cmu_snmp_parse(); 

/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/

static nk_snmp_var * _convertVariableFormat(pdu, varCount)
	struct snmp_pdu *pdu;
	int		*varCount;
{
	nk_snmp_var	*head,
			*curVar,
			*newVar;
	struct variable_list *oldVar;
	int		count,
			numVars;

	curVar = NULL;
	numVars = 0;
	oldVar = pdu->variables;
	while(oldVar)
	{
		numVars++;
		newVar = (nk_snmp_var *)malloc(sizeof(nk_snmp_var));
		if (curVar == NULL)
		{
			head = newVar;
		}
		else
		{
			curVar->next = newVar;
		}
		curVar = newVar;
		curVar->next = NULL;

		/* Text rep of the value */
		sprint_value(curVar->text_val, oldVar->name, 
			oldVar->name_length, oldVar);

		/* Copy the numeric OID in byte form */
		count = 0;
		while(count < oldVar->name_length)
		{
			curVar->oid[count] = oldVar->name_loc[count];
			count++;
		}
		curVar->oid_len = count;

		/* Get the variable name */
		sprint_objid(curVar->var, oldVar->name, oldVar->name_length);

		/* Store the value */
		switch(oldVar->type)
		{
			case ASN_INTEGER:
				curVar->int_val = *(int*)oldVar->val.integer;
				curVar->var_type = NK_SNMP_INT_TYPE;
				break;

			case ASN_COUNTER:
			case ASN_GAUGE:
			case ASN_TIMETICKS:
			case ASN_UINTEGER:
				curVar->uint_val = (u_int)oldVar->val.integer;
				curVar->var_type = NK_SNMP_UINT_TYPE;
				break;

			case ASN_COUNTER64:
				bcopy(oldVar->val.counter64, 
					&(curVar->int64_val),
					sizeof(curVar->int64_val));
				curVar->var_type = NK_SNMP_INT64_TYPE;
				break;

			case ASN_IPADDRESS:
				curVar->ipv4_val[0] = oldVar->val.string[0];
				curVar->ipv4_val[1] = oldVar->val.string[1];
				curVar->ipv4_val[2] = oldVar->val.string[2];
				curVar->ipv4_val[3] = oldVar->val.string[3];
				curVar->var_type = NK_SNMP_IPV4_TYPE;
				break;

			case ASN_OCTET_STR:
			case ASN_OPAQUE:
				if (oldVar->val_len)
					strncpy(curVar->text_val, 
						oldVar->val.string, 
						oldVar->val_len - 2);
				else
					*curVar->text_val = 0;
				curVar->var_type = NK_SNMP_OCTET_STR_TYPE;
				break;

			case ASN_OBJECT_ID:
				bcopy(oldVar->val.objid, curVar->byte_val, 
					oldVar->val_len);
				curVar->var_type = NK_SNMP_OID_TYPE;
				break;

			default:
				curVar->var_type = NK_SNMP_UNKNOWN_TYPE;
		}
		curVar->var_data_len = oldVar->val_len;


		oldVar = oldVar->next_variable;
	}
	*varCount = numVars;
	return(head);
}


static void _initLibrary()
{
	static	int	init = 0;

	if (init)
		return;

	ds_toggle_boolean(DS_LIBRARY_ID, DS_LIB_QUICK_PRINT);
	init++;
}


/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/


/*
** Public
** nkSnmpPrintError		- 
*/
void nkSnmpPrintError(fp)
	FILE	*fp;
{
	switch(errorCode)
	{
	case NK_SNMP_ERR_BAD_OID:
		fprintf(fp,"Bad or Unknown Object ID or Variable\n");
		break;
	case NK_SNMP_ERR_UNKNOWN_HOST:
		fprintf(fp,"Unknown host\n");
		break;
	}
	printf("%s\n",errorMsg);
}



char *nkSnmpGetError()
{
	static	char buf[1024];

	switch(errorCode)
	{
		case NK_SNMP_ERR_BAD_OID:
			sprintf(buf,
				"Bad or Unknown Object ID or Variable (%s)",
				errorMsg);
			break;
		case NK_SNMP_ERR_UNKNOWN_HOST:
			sprintf(buf, "Unknown host(%s)", errorMsg);
			break;
		default:
			sprintf(buf, "Error %d : %s", errorCode, errorMsg);
			break;
	}
	return(buf);
}

/*
** Public
** nkSnmpErrorCode		- Return the internal error code value
*/
int nkSnmpErrorCode()
{
	return(errorCode);
}



/*
** Public
** nkSnmpOpen			- Create an snmp session
*/
char *nkSnmpOpen(host, community)
	char	*host,
		*community;
{
	struct snmp_session session, *ss;

	/*
	* Initialize the SNMP library (theirs and ours)
	*/
	init_snmp("Network Knowledge");
	_initLibrary();

	/*
	* Initialize a "session" that defines who we're going to talk to
	*/
	snmp_sess_init( &session );                   
	session.version = SNMP_VERSION_1;
	session.peername = host;
	session.community = community;
	session.community_len = strlen(session.community);

	/*
	* Open the session
	*/
	SOCK_STARTUP;
	ss = snmp_open(&session); 
	if (ss == NULL && session.s_snmp_errno == SNMPERR_BAD_ADDRESS)
	{
		errorCode = NK_SNMP_ERR_UNKNOWN_HOST;
		strcpy(errorMsg,host);
	}
	return((char *)ss);
}


/*
** Public
** nkSnmpOpenTrap
*/

#ifdef NOT_DEF
char *nkSnmpOpenTrap()
{
	struct	snmp_session 	*session, *sptr;

	snmp_sess_init(session);
	session->peername = SNMP_DEFAULT_PEERNAME;
	session->version = SNMP_DEFAULT_VERSION;
	session->community_len = SNMP_DEFAULT_COMMUNITY_LEN;
	session->retries = SNMP_DEFAULT_RETRIES;
	session->timeout = SNMP_DEFAULT_TIMEOUT;
	session->local_port = SNMP_TRAP_PORT;
	session->callback = NULL;
	session->callback_magic = NULL;
	session->authenticator = NULL;

	SOCK_STARTUP;
	sptr = snmp_open(session);
	if (sptr == NULL)
	{
		SOCK_CLEANUP;
		return(NULL);
	}
	return((char *)sptr);
}
#endif

int nkSnmpTrapOpen()
{
	int	sock;
	struct	sockaddr_in addr;

	init_snmp("Network Knowledge");
	_initLibrary();
	sock = socket(AF_INET, SOCK_DGRAM,0);
	if (sock < 0)
		return(-1);
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(SNMP_TRAP_PORT);
	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) != 0)
	{
		return(-1);
	}
	return(sock);
}



nk_snmp_trap *nkSnmpTrapRead(sock)
	int	sock;
{
	int		numFds,
			res,
			len,
			fromLen;
	fd_set		fdset;
	char		*cp;
	static u_char 	packet[PACKET_LENGTH];
	nk_snmp_trap	*trap;
	struct hostent	*hp;
	struct snmp_pdu *pdu;
	struct snmp_session session;
	struct sockaddr_in from,
			*agent_addr;

	/*
	** Select on the sock and wait for some action
	** Note: this is a synch / blocking 'trap read'
	*/
	numFds = sock + 1;
	FD_ZERO(&fdset);
	FD_SET(sock, &fdset);
	res = select(numFds, &fdset, 0, 0, NULL);
	if (res < 0)
	{
		perror("select");
		return(NULL);
	}
	if (res == 0)
	{
		return(NULL);
	}
	if (! FD_ISSET(sock,&fdset))
	{
		return(NULL);
	}

	/*
	** Read the packet
	*/
	fromLen = sizeof(from);
	len = recvfrom(sock, (char *)packet, PACKET_LENGTH, 0,
		(struct sockaddr *)&from, &fromLen);
	if (len < 0)
	{
		perror("recvfrom");
		return(0);
	}

	/*
	** Do the "snmp" stuff on the data we read
	** We call the old cmu routine to fill in the pdu.  It appears
	** that the routines returns a malloced buffer that we must free
	** otherwise we'll bloat.  Must keep an eye on this to ensure
	** this doesn't change in a later version of the snmp library.
	*/
	snmp_sess_init(&session);
	pdu = snmp_pdu_create(0);
	bcopy(&from, &pdu->address, sizeof(from));
	pdu->reqid = 0;
	cp = cmu_snmp_parse(&session, pdu, packet, len); 
	if (cp)
		free(cp);

	/*
	** It had better be a trap!
	*/
	if (pdu->command != SNMP_MSG_TRAP && pdu->command != SNMP_MSG_TRAP2)
		return(NULL);

	/*
	** If it's an enterprise trap then store the info
	*/
	trap = (nk_snmp_trap *)malloc(sizeof(nk_snmp_trap));
	if (trap == NULL)
	{
		return(NULL);
	}
	if (pdu->enterprise_length != 0)
	{
		int	count = 0;

		while(count < pdu->enterprise_length)
		{
			trap->enterprise_oid[count] = pdu->enterprise[count];
			count++;
		}
		sprint_objid(trap->enterprise_var, pdu->enterprise, count);
		trap->enterprise_len = pdu->enterprise_length;
	}

	/*
	** Work out where the trap came from
	*/
	if (pdu->agent_addr.sa_family > 0)
		agent_addr = (struct sockaddr_in *)&(pdu->agent_addr);
	else
		agent_addr = (struct sockaddr_in *)&(pdu->address);

	cp = inet_ntoa(agent_addr->sin_addr);
	trap->source_addr[0] = atoi(cp);
	cp = index(cp,'.') + 1;
	trap->source_addr[1] = atoi(cp);
	cp = index(cp,'.') + 1;
	trap->source_addr[2] = atoi(cp);
	cp = index(cp,'.') + 1;
	trap->source_addr[3] = atoi(cp);

	hp = gethostbyaddr((char*) &agent_addr->sin_addr,
		sizeof(agent_addr->sin_addr), AF_INET);
	if (hp)
	{
		strcpy(trap->source_host, hp->h_name);
	}
	else
	{
		*trap->source_host = 0;
	}
	trap->variables = _convertVariableFormat(pdu,
		&trap->num_variables);

	/*
	** Fill in any other details
	*/
	trap->trap_type = pdu->trap_type;
	trap->enterprise_type = pdu->specific_type;

	snmp_free_pdu(pdu);
	return(trap);
}



/*
** Public
** nkSnmpClose			- Close off and free a session
*/
void nkSnmpClose(sessPtr)
	char	*sessPtr;
{
	if (sessPtr == NULL)
		return;
	snmp_close((struct snmp_session *)sessPtr);
	SOCK_CLEANUP;
	return;
}



/*
** Public
** nkSnmpFreePDU		- Free a request or response PDU
*/
void nkSnmpFreePDU(pdu)
	char	*pdu;
{
	snmp_free_pdu((struct snmp_pdu *)pdu);
	return;
}


/*
** Public
** nkSnmpFreeTrap		
*/
void nkSnmpFreeTrap(trap)
	nk_snmp_trap	*trap;
{
	nk_snmp_var	*curVar,
			*prevVar;

	curVar = trap->variables;
	while(curVar)
	{
		prevVar = curVar;
		curVar = curVar->next;
		free(prevVar);
	}
	free(trap);
}


/*
** Public
** nkSnmpCreatePDU		- Create an ampty PDU
*/
char *nkSnmpCreatePDU(op)
	int	op;
{
	struct 	snmp_pdu 	*request;
	int	nativeOp;

	switch(op)
	{
		case NK_SNMP_GET:
			nativeOp = SNMP_MSG_GET;
			break;
	
		case NK_SNMP_GETNEXT:
			nativeOp = SNMP_MSG_GETNEXT;
			break;
	
		case NK_SNMP_SET:
			nativeOp = SNMP_MSG_SET;
			break;
		default:
			errorCode = NK_SNMP_ERR_BAD_OP;
			return(NULL);
	}
	request = snmp_pdu_create(nativeOp);
	return((char *)request);
}


/*
** Public
** nkSnmpAddVariable		- Add a variable to a PDU (use with GET etc)
*/
int nkSnmpAddVariable(pduPtr, var, val)
	char	*pduPtr,
		*var,
		*val;
{
	oid 	curOid[MAX_OID_LEN];
	size_t 	curOidLen;
	struct	snmp_pdu	*pdu;
	char	*tmpStr;

	if (pduPtr == NULL)
	{
		errorCode = NK_SNMP_ERR_BAD_PDU;
		return(-1);
	}
	pdu = (struct snmp_pdu *)pduPtr;
	if (var == NULL)
	{
		errorCode = NK_SNMP_ERR_NO_OID;
		return(-1);
	}
	curOidLen = MAX_OID_LEN;
    	if (snmp_parse_oid(var, curOid, &curOidLen) == 0)
	{
		errorCode = NK_SNMP_ERR_BAD_OID;
		strcpy(errorMsg, var);
		return(-1);
	}
	if (val)
	{
    		if (snmp_add_var(pdu, curOid, curOidLen, '=', val))
		{
			tmpStr = snmp_api_errstring(snmp_errno);
			if (tmpStr != NULL)
			{
				sprintf(errorMsg, "%s - %s", var, tmpStr);
			}
			errorCode = NK_SNMP_ERR_BAD_OID;
			return(-1);
		}
	}
	else
	{
    		snmp_add_null_var(pdu, curOid, curOidLen);
	}
	return(0);
}

/*
** Public
** nkSnmpRequest		- Send the PDU to the host
*/

nk_snmp_result * nkSnmpRequest(sessPtr, pduPtr)
	char	*sessPtr,
		*pduPtr;
{
	int	result;
	struct 	snmp_session 	*session;
	struct 	snmp_pdu 	*request,
				*response;
	nk_snmp_result		*resPtr;

	if (sessPtr == NULL)
	{
		errorCode = NK_SNMP_ERR_SESSION;
		return(NULL);
	}
	else
	{
		session = (struct snmp_session *)sessPtr;
	}
	if (pduPtr == NULL)
	{
		errorCode = NK_SNMP_ERR_BAD_PDU;
		return(NULL);
	}
	else
	{
		request = (struct snmp_pdu *)pduPtr;
	}

	/*
	** 
	*/
	resPtr = (nk_snmp_result *)malloc(sizeof(nk_snmp_result));
	if (resPtr == NULL)
	{
		errorCode = NK_ERR_NO_MEMORY;
		return(NULL);
	}
	bzero(resPtr,sizeof(nk_snmp_result));
 
	/*
	** Send the Request and procss the response
	*/
	result = snmp_synch_response(session, request, &response);
	if (result == STAT_SUCCESS && response->errstat == SNMP_ERR_NOERROR) 
	{
		resPtr->result = NK_OK;
		resPtr->data = (void*)response;
		resPtr->variables = _convertVariableFormat(response,
			&resPtr->num_variables);
		return(resPtr);
	}
	if (result == STAT_TIMEOUT)
	{
		errorCode = NK_SNMP_ERR_TIMEOUT;
		strcpy(errorMsg, "SNMP operation timed out");
		return(NULL);
	}
	if (response)
	{
		switch(response->errstat)
		{
			case SNMP_ERR_NOSUCHNAME:
				errorCode = NK_SNMP_ERR_BAD_OID;
				break;
		}
		strcpy(errorMsg,snmp_errstring(response->errstat));
	}
	else
	{
		errorCode = NK_SNMP_ERR_GENERAL;
		strcpy(errorMsg,"Unknown SNMP error");
		*errorMsg = 0;
	}
	return(NULL);
}


int nkSnmpGetTrapType(pduPtr)
	char	*pduPtr;
{
	struct	snmp_pdu	*pdu;

	errorCode = NK_SNMP_ERR_OK;
	if (pduPtr == NULL)
	{
		errorCode = NK_SNMP_ERR_BAD_PDU;
		return(-1);
	}
	pdu = (struct snmp_pdu *)pduPtr;

	switch(pdu->trap_type) 
	{
  		case SNMP_TRAP_COLDSTART:
  			return(NK_SNMP_TRAP_COLDSTART);
		case SNMP_TRAP_WARMSTART:
			return(NK_SNMP_TRAP_WARMSTART);
		case SNMP_TRAP_LINKDOWN:
			return(NK_SNMP_TRAP_LINKDOWN);
		case SNMP_TRAP_LINKUP:
			return(NK_SNMP_TRAP_LINKUP);
		case SNMP_TRAP_EGPNEIGHBORLOSS:
			return(NK_SNMP_TRAP_EGPNEIGHBORLOSS);
		case SNMP_TRAP_ENTERPRISESPECIFIC:
			return(NK_SNMP_TRAP_ENTERPRISESPECIFIC);
  		default:
			return(-1);
	}
}

char *nkSnmpPrintOid(oid_ptr, len)
	char		*oid_ptr;
	int		len;
{
	static	char buf[1024],
		*cp;
	int	count;

	count = 0;
	cp = buf;
	while(count < len)
	{
		if (count > 0)
			*cp++ = '.';
		sprintf(cp,"%d", (u_char)oid_ptr[count]);
		cp = cp + strlen(cp);
		count++;
	}
	return(buf);
}
